#include "bookmoochercore.h"
#include "BMException.h"
#include "bookmoocherdata.h"
#include <QDir>
#include <qdesktopservices.h>
#include <qdebug.h>
#include <qnetworkreply.h>
#include <qurl.h>
#include <qfile.h>
#include <QtXml>
#include <QTextStream>
#include <QSettings>
#include <QStringList>
#include <QXmlStreamWriter>

BookMoocherCore::BookMoocherCore(QObject *parent) :
    QObject(parent), is_updating(false)
{
	// Inizializza i gestori di rete
	qnam = new QNetworkAccessManager(this);
	connect(qnam,SIGNAL(finished(QNetworkReply*)), this, SLOT(finishedBookmoochXMLFetch(QNetworkReply*)) );

	book_info_qnam = new QNetworkAccessManager(this);
	connect(book_info_qnam, SIGNAL(finished(QNetworkReply*)), this, SLOT(finishedBookInfoXMLFetch(QNetworkReply*)) );

	// Verifica l'esistenza e nel caso crea le directory dei dati
	QString data_dir = QDesktopServices::storageLocation(QDesktopServices::DataLocation);
	QDir dir;
	if (!dir.exists(data_dir)) {
		if (!dir.mkpath(data_dir)) throw BMException(tr("Unable to create data directory:") + data_dir);
	}

	/*
	// Si devono riferire alla stessa data
	QDateTime dt_l = QDateTime::currentDateTime();
	QDateTime dt_utc = dt_l.toUTC();
	int offset = 0;
	// Portland (dove c'è bookmooch) sta a -7 su UTC

	// Calcola la differenza dell'ora locale con UTC
	int foo_bar_baz = (dt_l.time().minute() * 60 + dt_l.time().hour() * 3600) - (dt_utc.time().minute() * 60 + dt_utc.time().hour() * 3600 );
	qDebug() << "Timezone diff:" << foo_bar_baz << " " << (foo_bar_baz / 3600);
	if (foo_bar_baz == 0) offset = -7;
	else if (foo_bar_baz > 0) {
		offset = -(foo_bar_baz / 3600) - 7;
	} else {
		foo_bar_baz = foo_bar_baz / 3600;
		if (foo_bar_baz < -7) offset = -foo_bar_baz + 7;
		else if (foo_bar_baz > -7) offset = -7 - foo_bar_baz;
		else offset = 0;
	}
	qDebug() << "Timezone offset:" << offset;
	*/
}

void BookMoocherCore::fetchBookInfo(const QString &ASIN) {
	qDebug() << "Fetching book info from: " + QString(CFG_BOOKMOOCH_API_BOOK_INFO_BY_ASINS) + QString("?asins=") + ASIN;
	book_info_qnam->get( QNetworkRequest(QUrl( QString(CFG_BOOKMOOCH_API_BOOK_INFO_BY_ASINS) + QString("?asins=") + ASIN )) );
}

void BookMoocherCore::fetchBookInfo(const BookListItem &item) {
	fetchBookInfo(item.ASIN);
}

/**
 * Esegue il fetch dei nuovi libri
 */
void BookMoocherCore::fetchBookmoochXMLFeed() {
	// Imposta il flag di aggiornamento
	is_updating = true;

	// lancia la lettura
	QString fetch_url(CFG_BOOKMOOCH_API_RECENT_URL);
	if (last_update_time.isValid()) fetch_url += QString("&since=%1").arg(last_update_time.toTime_t());

	qDebug() << "FETCHING FROM: " << QString(fetch_url);

	qnam->get(QNetworkRequest(QUrl( fetch_url )));
}

/*
 * Slot invocato quando termina l'operazione di fetch della sorgente
 * di dati XML
 * */
void BookMoocherCore::finishedBookmoochXMLFetch(QNetworkReply *reply) {
	last_network_reply_error = reply->error();
	BookList book_list;
	QString error_msg;
	if (reply->error() != QNetworkReply::NoError) {
		is_updating = false;
		emit finished(0, book_list, NetworkError, reply->errorString());
	} else {
		if (saveBookFeed(reply, error_msg)) {
			QFile last_feed_file( getLastFeedFilename() );
			if (last_feed_file.open(QIODevice::ReadOnly)) {
				int item_count = 0;
				if (!parseXML(&last_feed_file, item_count, book_list, error_msg)) {
					is_updating = false;
					emit finished(item_count, book_list, UpdateFailed, error_msg);
				} else {
					is_updating = false;
					emit finished(item_count, book_list, UpdateOK, error_msg);
				}
			} else {
				emit finished(0, book_list, UpdateFailed, last_feed_file.errorString());
			}
		} else {
			is_updating = false;
			emit finished(0, book_list, UpdateFailed, error_msg);
		}
	}
}

/**
  * Salva il feed XML dei dati
  *
  * Il feed XML in ingresso viene fornito nel parametro, in uscita viene utilizzato
  * il file ritornato da {@see getLastFeedFilename}.
  * @param QIODevice xml_source sorgente XML
  * @param QString error_msg eventuale messaggio di errore
  */
bool BookMoocherCore::saveBookFeed(QIODevice *xml_source, QString &error_msg) {
	// Salva il feed
	QFile last_feed_file( getLastFeedFilename()  );

	if (!last_feed_file.open(QIODevice::WriteOnly | QIODevice::Truncate | QIODevice::Text)) {
		error_msg = last_feed_file.errorString();
		return false;
	}

	QTextStream input_stream(xml_source);
	QTextStream output_stream(&last_feed_file);

	while (!input_stream.atEnd()) {
		output_stream << input_stream.readLine() << endl;
	}
	output_stream.flush();
	last_feed_file.close();
	return true;
}

/**
	Esegue il parsing di un feed RSS.
	Salva il feed e aggiorna l'elenco (filtrato) dell'elenco dei libri

	L'istanza di QIODevice deve essere aperta e leggibile.
	@return false se non è stato possibile procedere, true se tutto ok
*/
bool BookMoocherCore::parseXML(QIODevice *xml_source, int& new_elements_count, BookList &new_book_list, QString &error_msg) {
	new_elements_count = 0;
	error_msg.clear();

	if (!xml_source->isOpen() || !xml_source->isReadable() ) {
		error_msg = xml_source->errorString();
		return false;
	}

	// Esegue il processamento del feed in ingresso
	QDomDocument xml_doc;
	QString xml_err_msg;
	int xml_err_line = 0, xml_err_column = 0;
	bool parse_ok = xml_doc.setContent(xml_source, &xml_err_msg, &xml_err_line, &xml_err_column);
	if (parse_ok) {
		// Naviga i nodi
		QDomNodeList node_list = xml_doc.documentElement().elementsByTagName("asin");

		QSettings settings;
		QStringList countries = settings.value(CFG_COUNTRIES_LIST).toStringList();

		for (int ciclo = 0; ciclo < node_list.count(); ciclo++) {
			QDomNode item_node = node_list.item(ciclo);

			QDomNode node = item_node.firstChild();

			BookListItem feed_item;
			feed_item.is_new = true;

			while (!node.isNull()) {

				if (node.toElement().tagName() == "id") feed_item.ASIN = node.toElement().text();
				else if (node.toElement().tagName() == "Title") feed_item.title = node.toElement().text();
				else if (node.toElement().tagName() == "Author") feed_item.author = node.toElement().text();
				else if (node.toElement().tagName() == "added_when") {
					// feed_item.date = QLocale::c().toDateTime(node.toElement().text(),"ddd, dd MMM yyyy hh:mm:ss +0000");
					bool ok;
					feed_item.date = QDateTime::fromTime_t(QString( node.toElement().text() ).toUInt(&ok));
				} else if (node.toElement().tagName() == "added_where") feed_item.country = node.toElement().text();
				else if (node.toElement().tagName() == "added_by") feed_item.username = node.toElement().text();
				else if (node.toElement().tagName() == "ISBN") feed_item.ISBN = node.toElement().text();

				node = node.nextSibling();
			}
			if (!feed_item.ASIN.isEmpty()) {
				if (countries.isEmpty()) new_book_list.append(feed_item);
				else if (!feed_item.country.isEmpty()) {
					if (countries.contains(feed_item.country)) new_book_list.append(feed_item);
				}
			}
		}

		// Legge se esiste la lista dei libri salvata precedentemente e applica i filtri e/o aggiorna
		BookList old_book_list;

		if (QFile::exists( getBookListFilename() )) {
			// Se il caricamento della lista fallisce non può procedere
			QDateTime old_book_list_time;
			QString old_book_list_version;
			if (loadBookList( getBookListFilename(), old_book_list, old_book_list_time, old_book_list_version) ) {
				// Sincronizza le liste
				qDebug() << "BookMoocherCore::ParseXML: synchronizing lists";
				// Elimina dalla nuova lista dei libri i libri già presenti nella vecchia lista
				BookListMutableIterator new_book_list_iter(new_book_list);
				while(new_book_list_iter.hasNext()) {
					BookListItem nbi = new_book_list_iter.next();

					// Cerca l'elemento nella vecchia lista
					if (old_book_list.contains(nbi)) new_book_list_iter.remove();
				}
			} else {
				error_msg = tr("BookMoocherCore::parseXML: failed reading old book list.");
				qDebug() << "BookMoocherCore::parseXML: failed reading old book list.";
				return false;
			}
		} else {
			// Se non esiste allora procede comunque, dato che old_book_list
			// conterrà una lista vuota
		}


		// Conta i nuovi elementi
		new_elements_count = new_book_list.size();

		// Fonde le due liste vecchi + nuovi
		new_book_list.append(old_book_list);

		// Ordina e applica i filtri alla "nuova" lista
		applyFilters(new_book_list);

		// Infine salva, aggiornando la data dell'ultimo fetch
		last_update_time = QDateTime::currentDateTime();
		return saveBookList(getBookListFilename(), new_book_list, last_update_time, error_msg);
	} else {
		error_msg = tr("XML Parsing error: %1 at LINE: %2 COLUMN: %3").arg(xml_err_msg).arg(xml_err_line).arg(xml_err_column);
		qDebug() << "BookMoocherCore::parseXML: XML Parsing error: " << xml_err_msg << " LINE: " << xml_err_line << " COLUMN: " << xml_err_column;
		return false;
	}
}

void BookMoocherCore::applyFilters(BookList &book_list) {

	QSettings settings;

	QStringList countries = settings.value(CFG_COUNTRIES_LIST).toStringList();
	QStringList genres = settings.value(CFG_GENRES_LIST).toStringList();

	QDateTime dt;

	if (settings.contains(CFG_BOOK_LIST_MAX_BOOK_AGE)) {
		if (settings.value(CFG_BOOK_LIST_MAX_BOOK_AGE).canConvert(QVariant::Int)) {
			if (settings.value(CFG_BOOK_LIST_MAX_BOOK_AGE).toInt() > 0) dt = QDateTime::currentDateTime().addDays(-1 * settings.value(CFG_BOOK_LIST_MAX_BOOK_AGE).toInt() );
		}
	}

	int max_items = -1;
	if (settings.contains(CFG_BOOK_LIST_MAX_ITEMS)) {
		if (settings.value(CFG_BOOK_LIST_MAX_ITEMS).canConvert(QVariant::Int)) max_items = settings.value(CFG_BOOK_LIST_MAX_ITEMS).toInt();
	}

	applyFilters(book_list, countries, dt, max_items );
}

/**
 * Carica i dati della lista dei libri nel model
 *
*/
void BookMoocherCore::loadBookList(BookListModel *model) {
	BookList items;
	QString version;

	if (loadBookList(getBookListFilename(), items, last_update_time, version)) {
		model->setBookList(items);
		model->sort(0, Qt::DescendingOrder);
	}
}

/**
	Ritorna il path completo del file contenente l'ultimo feed eseguito.
*/
QString BookMoocherCore::getLastFeedFilename() {
	return QDesktopServices::storageLocation(QDesktopServices::DataLocation) + QDir::separator() + CFG_LAST_FEED_FILENAME;
}

/**
  Ritorna il path completo del file contenente la lista dei libri (filtrata) mostrata all'utente
  */
QString BookMoocherCore::getBookListFilename() {
	return QDesktopServices::storageLocation(QDesktopServices::DataLocation) + QDir::separator() + CFG_BOOK_LIST_FILENAME;
}

/**
 * Salva la lista in un file di testo
 */
bool BookMoocherCore::saveBookList(const BookList &book_list, const QDateTime &fetch_date, QString& error_msg) {
	if (is_updating) return false;

	return this->saveBookList(getBookListFilename(), book_list, fetch_date, error_msg);
}

bool BookMoocherCore::saveBookList(const QString &out_filepath, const BookList &book_list, const QDateTime &fetch_date, QString& error_msg) {
	error_msg.clear();

	QFile out_file( out_filepath );
	if (!out_file.open(QIODevice::ReadWrite | QIODevice::Truncate | QIODevice::Text)) {
		error_msg = out_file.errorString();
		qDebug() << "BookMoocherCore::saveBookList: Failed to save book list";
		return false;
	}

	QXmlStreamWriter xml(&out_file);

	xml.setAutoFormatting(true);
	xml.writeStartDocument("1.0");
	xml.writeStartElement("booklist");

	// Versione
	xml.writeEmptyElement("version");
	xml.writeAttribute("docver", "1.4");

	// Data di fetch
	xml.writeStartElement("fetch_date");
	if (fetch_date.isValid()) {
		xml.writeCharacters( QString::number(fetch_date.toTime_t()) );
	} else {
		xml.writeCharacters( QString::number(QDateTime::currentDateTime().toTime_t()) );
	}
	xml.writeEndElement();

	// I vari elementi
	BookListIterator iter(book_list);
	while (iter.hasNext()) {
		BookListItem item = iter.next();

		// -- APRE
		xml.writeStartElement("item");

		// titolo
		xml.writeTextElement("title", item.title);

		// author
		xml.writeTextElement("author", item.author);

		// nome utente
		xml.writeTextElement("username", item.username);

		// Data
		if (item.date.isValid()) {
			xml.writeTextElement("date", item.date.toString("dd-MM-yyyy hh:mm:ss"));
		}

		// ASIN
		xml.writeTextElement("asin", item.ASIN);

		// ISBN
		xml.writeTextElement("isbn", item.ISBN);

		// country
		xml.writeTextElement("country", item.country);

		// genres
		xml.writeTextElement("genres", item.genres.join(","));


		// -- CHIUDE
		xml.writeEndElement();

	}

	xml.writeEndElement();
	xml.writeEndElement();

	out_file.flush();
	out_file.close();

	return true;
}

/**
 * Carica la lista dei libri dal file XML nella lista
 * */
bool BookMoocherCore::loadBookList(const QString &in_filepath, BookList &book_list, QDateTime &list_update_time, QString &version) {
	qDebug() << "Loading from: " << in_filepath;
	book_list.clear();

	// Apre il file e lo processa
	QFile input_file(in_filepath);
	// Verifica prima l'esistenza del file coi dati
	if (!input_file.exists()) {
		qDebug() << tr("Loading failed: file doesn't exists.");
		return false;
	}
	if (!input_file.open(QIODevice::ReadOnly | QIODevice::Text)) {
		qDebug() << tr("Loading failed: can't open file.");
		return false;
	}

	QDomDocument xml;
	int err_line, err_col;
	QString err_msg;
	list_update_time = QDateTime();
	version = "";
	if (xml.setContent(&input_file, &err_msg, &err_line, &err_col)) {

		// Legge versione e data fetch
		QDomNode n = xml.documentElement();
		if (n.nodeName() == "booklist") {
			n = n.firstChild();

			while (!n.isNull()) {
				if (n.nodeName() == "version") {
					if (n.hasAttributes()) {
						if (n.attributes().contains("docver")) {
							version = n.attributes().namedItem("docver").nodeValue();
						}
					}
				} else if (n.nodeName() == "fetch_date") {
					// Processa la data dell'ultimo fetch
					if (n.hasChildNodes() && n.firstChild().isText()) {
						bool ok;
						uint lut = n.firstChild().toText().data().toUInt(&ok);
						if (ok) list_update_time.setTime_t(lut);
					}

					qDebug() << "Data ultimo fetch: " << (last_update_time.isValid() ? last_update_time.toString("dd-MM-yyyy hh:mm:ss") : "N/A" );
				} else if (n.nodeName() == "item") {
					// Processa un elemento
					BookListItem item;

					QDomNode sub_node = n.firstChild();

					while (!sub_node.isNull()) {
						if ( sub_node.nodeName() == "title" ) {
							if (!sub_node.firstChild().isNull() && sub_node.firstChild().isText())
								item.title = sub_node.firstChild().toText().data();
						} else if ( sub_node.nodeName() == "author" ) {
							if (!sub_node.firstChild().isNull() && sub_node.firstChild().isText())
								item.author = sub_node.firstChild().toText().data();
						} if ( sub_node.nodeName() == "isbn" ) {
							if (!sub_node.firstChild().isNull() && sub_node.firstChild().isText())
								item.ISBN = sub_node.firstChild().toText().data();
						} else if (sub_node.nodeName() == "username") {
							if (!sub_node.firstChild().isNull() && sub_node.firstChild().isText())
								item.username = sub_node.firstChild().toText().data();
						} else if (sub_node.nodeName() == "date") {
							if (!sub_node.firstChild().isNull() && sub_node.firstChild().isText()) {
								item.date = QDateTime::fromString(sub_node.firstChild().toText().data(), "dd-MM-yyyy hh:mm:ss");
							}
						} else if (sub_node.nodeName() == "asin") {
							if (!sub_node.firstChild().isNull() && sub_node.firstChild().isText())
								item.ASIN = sub_node.firstChild().toText().data();
						} else if (sub_node.nodeName() == "country") {
							if (!sub_node.firstChild().isNull() && sub_node.firstChild().isText())
								item.country = sub_node.firstChild().toText().data();
						} else if (sub_node.nodeName() == "genres") {
							if (!sub_node.firstChild().isNull() && sub_node.firstChild().isText()) {
								// item.genres = sub_node.firstChild().toText().data().split(",", QString::SkipEmptyParts);
							}
						}

						sub_node = sub_node.nextSibling();
					}

					book_list.append(item);
				}

				n = n.nextSibling();
			}

		} else {
			qDebug() << "Invalid input XML";
			return false;
		}
	} else {
		qDebug() << "Loading book list failed: " << err_msg << " LINE: " << err_line << " COL: " << err_col << endl;
		return false;
	}

	input_file.close();
	return true;
}

void BookMoocherCore::applyFilters(BookList &book_list, QStringList &countries, QDateTime &remove_before, int limit_entries, Qt::SortOrder order) {
	if (book_list.isEmpty()) return;

	// Filtra per città e data
	BookListMutableIterator iter(book_list);
	bool should_delete = false;
	while (iter.hasNext()) {
		BookListItem item = iter.next();

		if (!countries.isEmpty()) {
			should_delete = !countries.contains(item.country);
		}
		if (!should_delete && remove_before.isValid()) {
			should_delete = (item.date < remove_before);
		}

		if (should_delete) iter.remove();
	}

	// Ordina i restanti elementi per data
	if (!book_list.isEmpty()) {
		// Ordina
		if (order == Qt::AscendingOrder) qSort(book_list);
		else qSort(book_list.begin(), book_list.end(), qGreater<BookListItem>());

		// Rimuove gli elementi
		if (limit_entries > 0 && book_list.size() > limit_entries) {
			// Salta gli elementi "buoni"
			iter.toFront();
			while (iter.hasNext() && limit_entries) {
				limit_entries--;
				iter.next();
			}

			// toglie quelli non voluti
			while(iter.hasNext()) {
				iter.next();
				iter.remove();
			}
		}
	}
}

bool BookMoocherCore::convertBookListToLatestVersion(const QString &in_filepath, BookList &book_list, QString &error_msg) {
	book_list.clear();
	error_msg.clear();

//	if (!QFile::exists(in_filepath)) {
//		qDebug() << "BookMoocherCore::convertBookListToLatestVersion: input file doesn't exists.";
//		return false;
//	}

	QFile input_file(in_filepath);
	if (!input_file.open(QIODevice::ReadOnly | QIODevice::Text)) {
		qDebug() << "BookMoocherCore::convertBookListToLatestVersion: can't open input file: " << in_filepath;
		error_msg = input_file.errorString();
		return false;
	}

	// Legge il file XML
	QDomDocument xml;
	QString err_msg; int err_line, err_col;
	if (xml.setContent(&input_file, &err_msg, &err_line, &err_col)) {
		QDomNode root_node = xml.documentElement();
		if (root_node.nodeName() == "booklist") {
			int fetch_date = 0;
			QString version;
			QDomNode node = root_node.firstChild();
			while (!node.isNull()) {
				if (node.nodeName() == "version") {
					if (node.attributes().contains("docver")) version = node.attributes().namedItem("docver").nodeValue();
				} else if (node.nodeName() == "fetch_date") {
					// Le versioni precedenti alla 1.4 contengono un bug che replica il nodo "fetch_date"
					// prende perciò quello col valore maggiore
					bool ok_fd;
					int fd = node.nodeValue().toInt(&ok_fd);
					if (ok_fd) {
						if (fd > fetch_date) fetch_date = fd;
					}
				} else if (node.nodeName() == "item") {
					// Processa un elemento
					BookListItem item;
					QDomNode item_node = node.firstChild();
					while (!item_node.isNull()) {
						if (item_node.nodeName() == "title") {
							if (item_node.hasChildNodes()) item.title = item_node.firstChild().toText().data();
						} else if (item_node.nodeName() == "author") {
							if (item_node.hasChildNodes()) item.author = item_node.firstChild().toText().data();
						} else if (item_node.nodeName() == "username") {
							if (item_node.hasChildNodes()) item.username = item_node.firstChild().toText().data();
						} else if (item_node.nodeName() == "asin") {
							if (item_node.hasChildNodes()) item.ASIN = item_node.firstChild().toText().data();
						} else if (item_node.nodeName() == "isbn") {
							if (item_node.hasChildNodes()) item.ISBN = item_node.firstChild().toText().data();
						} else if (item_node.nodeName() == "date") {
							if (item_node.hasChildNodes()) item.date = QDateTime::fromString(item_node.firstChild().toText().data(), "dd-MM-yyyy hh:mm:ss");
						} else if (item_node.nodeName() == "country") {
							if (item_node.hasChildNodes())
								if (BookMoocherData::getCountries().contains( item_node.firstChild().toText().data() ))
									item.country = item_node.firstChild().toText().data();
						}

						item_node = item_node.nextSibling();
					}

					if (!item.ASIN.isEmpty() && !item.title.isEmpty() && item.date.isValid() && !item.country.isEmpty()) {
						book_list.append(item);
					} else {
						qDebug() << QString("ITEM DISCARDED - ASIN: %1 ISBN: %2 TITLE: \"%3\" DATE: %4 COUNTRY: %5 ").arg(item.ASIN).arg(item.ISBN).arg(item.title).arg(item.date.isValid() ? "V" : "N").arg(item.country);
					}
				}

				node = node.nextSibling();
			}

			// Ha ottenuto una lista, la salva
			input_file.close();
			QDateTime fd = (fetch_date == 0 ? QDateTime::currentDateTime() : QDateTime::fromTime_t(fetch_date));

			QSettings settings;

			applyFilters(book_list);

			return saveBookList(in_filepath, book_list, fd, error_msg);
		} else {
			input_file.close();
			qDebug() << "BookMoocherCore::convertBookListToLatestVersion: input file isn't valid.";
			error_msg = tr("BookMoocherCore::convertBookListToLatestVersion: input file isn't valid.");
			return false;
		}

		return true;
	} else {
		input_file.close();
		error_msg = tr("BookMoocherCore::convertBookListToLatestVersion: input file parsing error: %1 at LINE: %2 COL: %3").arg(err_msg).arg(err_line).arg(err_col);
		qDebug() << "BookMoocherCore::convertBookListToLatestVersion: input file parsing error: " << err_msg << " line: " << err_line << " col: " << err_col;
		return false;
	}
}

bool BookMoocherCore::convertBookListToLatestVersion(BookListModel *model, QString &error_msg) {
	BookList bl;

	if (convertBookListToLatestVersion(getBookListFilename(), bl, error_msg)) {
		model->setBookList(bl);
		model->sort(0, Qt::DescendingOrder);
		return true;
	}
	return false;
}

const QDateTime &BookMoocherCore::getLastUpdateTime() const {
	return last_update_time;
}

void BookMoocherCore::finishedBookInfoXMLFetch(QNetworkReply *reply) {
	BookInfoItem book_info;
	if (reply->error() != QNetworkReply::NoError) {
		emit finishedBookInfoFetch(book_info, NetworkError, reply->errorString());
	} else {
		QString err_msg;
		if (parseBookInfoXML(reply, &book_info, err_msg)) {
			emit finishedBookInfoFetch(book_info, UpdateOK, err_msg);
		} else {
			emit finishedBookInfoFetch(book_info, UpdateFailed, err_msg);
		}
	}
}

bool BookMoocherCore::parseBookInfoXML(QIODevice *source, BookInfoItem *book_info, QString &err_msg) {
	QDomDocument xml;
	int err_col, err_line;
	if (xml.setContent(source, &err_msg, &err_line, &err_col)) {
		// Estrae la radice
		QDomElement root = xml.documentElement();
		if (root.nodeName() == "asins") {
			// Scorre il primo nodo, l'unico che interessi, per adesso
			if (root.hasChildNodes()) {
				// Verifica se ci sia stato un errore
				QDomNode result_code = root.firstChild().namedItem("result_code");
				if (result_code.isNull()) {
					book_info->ASIN.clear();
					QDomNodeList node_list = root.firstChild().childNodes();
					for (int c = 0; c < node_list.size(); c++) {
						QDomNode n = node_list.at(c);
						if (n.nodeName() == "id") book_info->ASIN = n.firstChild().toText().data();
						else if (n.nodeName() == "Author") book_info->author = n.firstChild().toText().data();
						else if (n.nodeName() == "Binding") book_info->binding = n.firstChild().toText().data();
						else if (n.nodeName() == "DetailPageURL") book_info->detail_page_URL = n.firstChild().toText().data();
						else if (n.nodeName() == "Edition") book_info->edition = n.firstChild().toText().data();
						else if (n.nodeName() == "ISBN") book_info->ISBN = n.firstChild().toText().data();
						else if (n.nodeName() == "NumberOfPages") book_info->number_of_pages = n.firstChild().toText().data().toInt();
						else if (n.nodeName() == "Publisher") book_info->publisher = n.firstChild().toText().data();
						else if (n.nodeName() == "Title") book_info->title = n.firstChild().toText().data();
						else if (n.nodeName() == "store") book_info->store = n.firstChild().toText().data();
						else if (n.nodeName() == "PublicationDate") book_info->publication_date = QDateTime::fromString(n.firstChild().toText().data(), "yyyy-MM-dd");
					}
					// A questo punto se non abbiamo almeno l'ASIN valido, è tutto inutile
					if (book_info->ASIN.isEmpty()) {
						err_msg = tr("Invalid book info source XML");
					} else {
						return true;
					}
				} else {
					// C'è un errore, prende il messaggio di errore
					QDomNode result_text = root.firstChild().namedItem("result_text");
					err_msg = result_text.firstChild().toText().data();
				}
			} else {
				err_msg = tr("Invalid book info source XML");
			}
		} else {
			err_msg = tr("Invalid book info source XML");
		}
	} else {
		err_msg += " LINE: " + QString::number(err_line) + " COL: " + QString::number(err_col);
	}

	return false;
}
